/*
 * query.cpp
 *
 *  Created on: 2009/10/09
 *      Author: shu
 */

#include "aligner_parameters.h"
#include "sequence_reader.h"
#include "sequence.h"
#include "encoder.h"
#include "dna_sequence.h"
#include "dna_type.h"
#include "protein_type.h"
#include "statistics.h"
#include "translator.h"
#include "translated_query.h"
#include "normal_query.h"
#include "queries.h"
#include "kmer_counter.h"
#include <cmath>
#include <cstddef>
#include <algorithm>
#include <numeric>
#include <typeinfo>
#include <string>
#include <stdint.h>
#include <vector>
#include <iostream>
#include <stdexcept>
#include <utility>
#include <functional>

using namespace std;

namespace
{

double entropy(vector<unsigned int> &cnt)
{
  const double pseudocount = 1e-5;
  double sum = accumulate(cnt.begin(), cnt.end(), 0) + pseudocount * cnt.size();
  double ent = 0.0;
  for (vector<unsigned int>::iterator i=cnt.begin();
      i<cnt.end(); ++i) {
    double x = (*i + pseudocount) / sum;
    ent += x * log(x);
  }
  return -ent;
}

}



Queries::Queries(const AlignerParameters &parameters) :
  parameters_(parameters),
  max_sequence_length_(0),
  reader_(CreateSequenceReader(parameters.GetQueriesFileNmae().c_str(), FASTA)),
  next_sequence_(SequencePtr())
{
  Next();
}

void Queries::Next() {
  Clear();
  max_sequence_length_ = 0;
  uint32_t sum_length= 0;
  int i=0;

  if(!next_sequence_) {
    next_sequence_ = reader_->Read();
  }

  if (!next_sequence_) {
    return;
  }

  uint32_t max_sequences_length = parameters_.GetMaxQueriesLength();

  while(GetLength(next_sequence_) <= 0) {
    next_sequence_ = reader_->Read();
    if (!next_sequence_) {
      return;
    }
  }

  if (GetLength(next_sequence_) >= max_sequences_length) {
    throw invalid_argument("the max queries length is too short.");
  }

  vector<pair<double, int> > entropies_indexes;
  vector<uint32_t> sequence_lengths;
  DnaType dt;
  KmerCounter kc(dt.GetRegularLetters());
  const int kmer_size = 3;
  vector<unsigned int> cnt(kc.kmer_type_size(kmer_size));
  while (next_sequence_) {
    if (GetLength(next_sequence_) > 0) {
      if (sum_length + GetLength(next_sequence_) >= max_sequences_length) {
        break;
      }
      sum_length += GetLength(next_sequence_);
      sequence_lengths.push_back(GetLength(next_sequence_));
      if (max_sequence_length_ < GetLength(next_sequence_)) {
        max_sequence_length_ = GetLength(next_sequence_);
      }
      SetQuery(next_sequence_);
      fill(cnt.begin(), cnt.end(), 0);
      kc.count(next_sequence_->GetSequenceData(), kmer_size, cnt);
      entropies_indexes.push_back(make_pair(entropy(cnt), i));
      ++i;
    }
    next_sequence_ = reader_->Read();
  }

  if ((parameters_.GetEntropyThresholdForReordering() > 0.0) ||
      (parameters_.GetEntropyThresholdForFiltering() > 0.0)) {
    FilterAndReorder(entropies_indexes, sequence_lengths);
  }
}

uint32_t Queries::GetLength(SequencePtr sequence) {
  if (parameters_.Translated()) {
    return TranslatedQuery::GetSequenceLength(sequence->GetSequenceData().size());
  } else {
    return NormalQuery::GetSequenceLength(sequence->GetSequenceData().size());
  }
}

void Queries::SetQuery(SequencePtr sequence) {
  if (parameters_.Translated()) {
    queries_.push_back(new TranslatedQuery(*sequence, parameters_));
  } else {
    queries_.push_back(new NormalQuery(*sequence, parameters_));
  }
}

void Queries::Clear() {
  for (vector<Query *>::iterator it = queries_.begin(); it != queries_.end(); ++it) {
    delete *it;
  }
  queries_.clear();
}

// filter the low-complexity sequences (LCSs), and
// reorder the queries in such a way that all LCSs, which have low entropy,
// precede other sequences.
void Queries::FilterAndReorder(vector<pair<double, int> >& entropies_indexes,
    vector<uint32_t>& sequence_lengths)
{
  typedef pair<double, int> pdi;
  // filtering
  if (parameters_.GetEntropyThresholdForFiltering() > 0.0) {
    const pdi filter_threshold(parameters_.GetEntropyThresholdForFiltering(), 0);
    vector<pdi>::iterator mid = partition(entropies_indexes.begin(),
        entropies_indexes.end(), bind2nd(greater_equal<pdi>(), filter_threshold));
    for (vector<pdi>::iterator i = mid; i != entropies_indexes.end(); ++i) {
      delete queries_[i->second];
      queries_[i->second] = 0;
    }
    max_sequence_length_ = 0;
    for (vector<pdi>::iterator i = entropies_indexes.begin(); i != mid; ++i) {
      if (sequence_lengths[i->second] > max_sequence_length_) {
        max_sequence_length_ = sequence_lengths[i->second];
      }
    }
    entropies_indexes.erase(mid, entropies_indexes.end());
  }
  // reordering
  if (parameters_.GetEntropyThresholdForReordering() > 0.0 &&
      parameters_.GetEntropyThresholdForReordering() > parameters_.GetEntropyThresholdForFiltering()) {
    const pdi reorder_threshold(parameters_.GetEntropyThresholdForReordering(), 0);
    partition(entropies_indexes.begin(), entropies_indexes.end(), bind2nd(less<pdi>(), reorder_threshold));
  }
  // reflect the filtering and reordering operation
  vector<Query *> q(entropies_indexes.size());
  vector<uint32_t> sl(entropies_indexes.size());
  for (size_t i=0; i<entropies_indexes.size(); ++i) {
    q[i] = queries_[entropies_indexes[i].second];
    sl[i] = sequence_lengths[entropies_indexes[i].second];
    entropies_indexes[i].second = i;
  }
  queries_.swap(q);
  sequence_lengths.swap(sl);
}
